(***********************************************************************)
(*                                                                     *)
(*                        OCamlP3l -- Example                          *)
(*                                                                     *)
(*             projets Estime & Cristal, INRIA-Rocquencourt            *)
(*                                                                     *)
(*  Copyright 2003-2007 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: my_data.ml,v 1.1 2007/01/17 23:36:00 fclement Exp $ *)

open Config;;
open Rdec2;;

type subdomain =
    {config_filename : string;
     config : configuration;
     rdec : regular_decomposition;
     num : int;
     iter : int;
     bd : boundary;
     err : float * float};;

let zero_length_error () =
  invalid_arg "My_data: empty stream item"
and constant_field_error s =
  failwith (Printf.sprintf "My_data: field %s should be constant" s)
and same_length_error () =
  invalid_arg "My_data: inconsistent lengths";;

let check_length v =
  let n = Array.length v in
  if n = 0 then zero_length_error ();;

(* Extraction functions *)

let get_unique get s v =
  check_length v;
  let res = get v.(0) in
  if My_array.for_all (fun x -> get x = res) v then res else
  constant_field_error s;;

let get_config_filename v =
  get_unique (fun subd -> subd.config_filename) "config_filename" v
and get_config v = get_unique (fun subd -> subd.config) "config" v
and get_rdec v = get_unique (fun subd -> subd.rdec) "rdec" v
and get_nums v = Array.map (fun subd -> subd.num) v
and get_iter v = get_unique (fun subd -> subd.iter) "iter" v
and get_bds v = Array.map (fun subd -> subd.bd) v
and get_errs v = Array.map (fun subd -> subd.err) v;;

(* Insertion function *)

let set_bds bds v =
  My_array.array_seti (fun i subd -> {subd with bd = bds.(i)}) v;;
